//          MT5.

#define TOSTRING(A)  #A + " = " + (string)(A) + "\n"
#define TOSTRING2(A) #A + " = " + EnumToString(A) + " (" + (string)(A) + ")\n"

string ToString( const MqlTradeRequest &Request )
{
  return(TOSTRING2(Request.action) + TOSTRING(Request.magic) + TOSTRING(Request.order) +
         TOSTRING(Request.symbol) + TOSTRING(Request.volume) + TOSTRING(Request.price) +
         TOSTRING(Request.stoplimit) + TOSTRING(Request.sl) +  TOSTRING(Request.tp) +
         TOSTRING(Request.deviation) + TOSTRING2(Request.type) + TOSTRING2(Request.type_filling) +
         TOSTRING2(Request.type_time) + TOSTRING(Request.expiration) + TOSTRING(Request.comment) +
         TOSTRING(Request.position) + TOSTRING(Request.position_by));
}

string ToString( const MqlTradeResult &Result )
{
  return(TOSTRING(Result.retcode) + TOSTRING(Result.deal) + TOSTRING(Result.order) +
         TOSTRING(Result.volume) + TOSTRING(Result.price) + TOSTRING(Result.bid) +
         TOSTRING(Result.ask) + TOSTRING(Result.comment) + TOSTRING(Result.request_id) +
         TOSTRING(Result.retcode_external));
}

string ToString( const MqlTradeCheckResult &CheckResult )
{
  return(TOSTRING(CheckResult.retcode) + /*TOSTRING(CheckResult.balance) + TOSTRING(CheckResult.equity) +
         TOSTRING(CheckResult.profit) + TOSTRING(CheckResult.margin) + TOSTRING(CheckResult.margin_free) +
         TOSTRING(CheckResult.margin_level) +*/ TOSTRING(CheckResult.comment));
}

string TimeToString( const long Value )
{
  return((string)(datetime)(Value / 1000) + "." + (string)IntegerToString(Value % 1000, 3, '0'));
}

#define TOSTRING3(A) #A + " = " + TimeToString(A) + "\n"

string OrderToString( void )
{
  return(TOSTRING(OrderTicket()) + TOSTRING3(OrderOpenTimeMsc()) + TOSTRING2((ENUM_ORDER_TYPE)OrderType()) + TOSTRING(OrderLots()) + TOSTRING(OrderSymbol()) +
         TOSTRING(OrderOpenPrice()) + TOSTRING(OrderStopLoss()) + TOSTRING(OrderTakeProfit()) + TOSTRING3(OrderCloseTimeMsc()) + TOSTRING(OrderClosePrice()) +
         TOSTRING(OrderProfit()));
}

string GetTickFlag( uint tickflag )
{
  string flag = NULL;

#define TICKFLAG_MACRO(A) flag += ((bool)(tickflag & TICK_FLAG_##A)) ? " TICK_FLAG_" + #A : ""; \
                          tickflag -= tickflag & TICK_FLAG_##A;
  TICKFLAG_MACRO(BID)
  TICKFLAG_MACRO(ASK)
  TICKFLAG_MACRO(LAST)
  TICKFLAG_MACRO(VOLUME)
  TICKFLAG_MACRO(BUY)
  TICKFLAG_MACRO(SELL)
#undef TICKFLAG_MACRO

  if (tickflag)
    flag += " FLAG_UNKNOWN (" + (string)tickflag + ")";

  return(flag);
}

string ToString( const MqlTick &Tick )
{
  return(TOSTRING3(Tick.time_msc) + TOSTRING(Tick.bid) + TOSTRING(Tick.ask) + TOSTRING(Tick.last)+ TOSTRING(Tick.volume) + TOSTRING(Tick.flags) + GetTickFlag(Tick.flags) + "\n");
}

class FILE
{
public:
  const string Name;
  const int handle;

#define _H(A) if (this.handle != INVALID_HANDLE) A

  FILE( const string FileName ) : Name(FileName), handle(FileOpen(this.Name, FILE_READ | FILE_WRITE | FILE_IS_TEXT | FILE_SHARE_READ))
  {
    _H(FileSeek(this.handle, 0, SEEK_END));
  }

  ~FILE() { _H(FileClose(this.handle)); }

  uint Write( const string Str ) const
  {
    return(FileWrite(this.handle, Str));
  }
};

ENUM_DAY_OF_WEEK GetDayOfWeek( const datetime time )
{
  MqlDateTime sTime = {0};

  ::TimeToStruct(time, sTime);

  return((ENUM_DAY_OF_WEEK)sTime.day_of_week);
}

bool SessionTrade( const string Symb )
{
  datetime TimeNow = ::TimeTradeServer();

  const ENUM_DAY_OF_WEEK DayOfWeek = GetDayOfWeek(TimeNow);

  TimeNow %= 24 * 60 * 60;

  bool Res = false;
  datetime From, To;

  for (int i = 0; (!Res) && ::SymbolInfoSessionTrade(Symb, DayOfWeek, i, From, To); i++)
    Res = ((From <= TimeNow) && (TimeNow < To));

  return(Res);
}

bool SymbolTrade( const string Symb, const int TimeOut = INT_MAX )
{
  MqlTick Tick;

  return(::SymbolInfoTick(Symb, Tick) ? ((Tick.bid != 0) && (Tick.ask != 0) && SessionTrade(Symb) &&
         (TimeCurrent() - Tick.time < TimeOut) && // https://www.mql5.com/ru/forum/1111/page2237#comment_8159562
         ((ENUM_SYMBOL_TRADE_MODE)::SymbolInfoInteger(Symb, SYMBOL_TRADE_MODE) == SYMBOL_TRADE_MODE_FULL)) : false);
}

#define MT4ORDERS_BENCHMARK Alert(MT4ORDERS::LastTradeRequest.symbol + " " +       \
                                  (string)MT4ORDERS::LastTradeResult.order + " " + \
                                  MT4ORDERS::LastTradeResult.comment);             \
                            Print(ToString(MT4ORDERS::LastTradeRequest) +          \
                                  ToString(MT4ORDERS::LastTradeResult));

#include <MT4Orders.mqh> // https://www.mql5.com/ru/code/16006

#define Point (SymbolInfoDouble(Symb, SYMBOL_POINT))
#define Digits ((int)SymbolInfoInteger(Symb, SYMBOL_DIGITS))
#define Bid (SymbolInfoDouble(Symb, SYMBOL_BID))
#define Ask (SymbolInfoDouble(Symb, SYMBOL_ASK))
#define Last (SymbolInfoDouble(Symb, SYMBOL_LAST))

#define ERROR_LOG(A)                                                                                                                                  \
  {                                                                                                                                                   \
      MqlTick CurrTick, PrevTick;                                                                                                                     \
                                                                                                                                                      \
      SymbolInfoTick(Symb, PrevTick);                                                                                                                 \
      Ticket = A;                                                                                                                                     \
      SymbolInfoTick(Symb, CurrTick);                                                                                                                 \
                                                                                                                                                      \
      if ((Ticket <= 0) && (MT4ORDERS::LastTradeResult.retcode >= TRADE_RETCODE_DONE) &&                                                              \
                           (MT4ORDERS::LastTradeResult.retcode != TRADE_RETCODE_PRICE_OFF))                                                           \
      {                                                                                                                                               \
        const string Str = TOSTRING(TimeCurrent()) + #A + "\n" +                                                                                      \
                           ToString(MT4ORDERS::LastTradeRequest) + ToString(MT4ORDERS::LastTradeResult) + ToString(MT4ORDERS::LastTradeCheckResult) + \
                           (!Ticket ? OrderToString() : NULL) +                                                                                       \
                           TOSTRING(Ticket) +                                                                                                         \
                           "PrevTick:\n" + ToString(PrevTick) + "CurrentTick:\n" + ToString(CurrTick) +                                               \
                           TOSTRING((int)(TimeCurrent() - CurrTick.time)) + "\n" +                                                                    \
                           TOSTRING(SymbolInfoInteger(Symb, SYMBOL_TRADE_STOPS_LEVEL)) +                                                              \
                           TOSTRING(SymbolInfoInteger(Symb, SYMBOL_TRADE_FREEZE_LEVEL)) +                                                             \
                           TOSTRING(SymbolInfoDouble(Symb, SYMBOL_TRADE_TICK_SIZE)) +                                                                 \
                           TOSTRING(SymbolInfoDouble(Symb, SYMBOL_SESSION_PRICE_LIMIT_MIN)) +                                                         \
                           TOSTRING(AccountInfoString(ACCOUNT_SERVER)) + TOSTRING(TerminalInfoInteger(TERMINAL_PING_LAST));                           \
                                                                                                                                                      \
        Print(Str);                                                                                                                                   \
        File.Write(Str);                                                                                                                              \
                                                                                                                                                      \
        if (MT4ORDERS::LastTradeResult.retcode == TRADE_RETCODE_DONE)                                                                                 \
        {                                                                                                                                             \
          Alert(TOSTRING(MT4ORDERS::LastTradeResult.retcode) + TOSTRING(Ticket));                                                                     \
                                                                                                                                                      \
          ExpertRemove();                                                                                                                             \
                                                                                                                                                      \
          return;                                                                                                                                     \
        }                                                                                                                                             \
      }                                                                                                                                               \
  }

double CorrectPrice( const string &Symb, const double Price )
{
  const double TickSize = MathMax(SymbolInfoDouble(Symb, SYMBOL_TRADE_TICK_SIZE), Point);

  return(NormalizeDouble(((int)(Price / TickSize + 0.1)) * TickSize, Digits));
}

double GetMinLot( const string &Symb )
{
  const double StepLot = SymbolInfoDouble(Symb, SYMBOL_VOLUME_STEP);
  const double MinLot = SymbolInfoDouble(Symb, SYMBOL_VOLUME_MIN);

  double Res = MathMax(SymbolInfoDouble(Symb, SYMBOL_VOLUME_MIN), StepLot);

  if (!Res)
    Res = 0.01;
  else if (StepLot)
  {
    Res = NormalizeDouble(((int)(Res / StepLot + 0.1)) * StepLot, 2);

    if (Res < MinLot)
      Res = NormalizeDouble(Res + StepLot, 2);
  }

  return(Res);
}

void OrderSendTest( string Symb = NULL, uint LimitOffset = 1000 )
{
  static FILE File(__FUNCTION__ + ".txt");

  if (!AccountInfoInteger(ACCOUNT_TRADE_EXPERT) || !AccountInfoInteger(ACCOUNT_TRADE_ALLOWED) || !TerminalInfoInteger(TERMINAL_TRADE_ALLOWED))
  {
    MessageBox(TOSTRING(AccountInfoInteger(ACCOUNT_TRADE_EXPERT)) +
               TOSTRING(AccountInfoInteger(ACCOUNT_TRADE_ALLOWED)) +
               TOSTRING(TerminalInfoInteger(TERMINAL_TRADE_ALLOWED)));

    ExpertRemove();

    return;
  }

  if (Symb == NULL)
    Symb = _Symbol;

//  if (SymbolTrade(Symb, 55))
  if (SymbolTrade(Symb))
  {
    const bool DemoAccount = ((ENUM_ACCOUNT_TRADE_MODE)AccountInfoInteger(ACCOUNT_TRADE_MODE) == ACCOUNT_TRADE_MODE_DEMO);

    const double MinLot = GetMinLot(Symb);
    const uint TickSize = MathMax((uint)(SymbolInfoDouble(Symb, SYMBOL_TRADE_TICK_SIZE) / Point + 0.1), 1);

    LimitOffset -= LimitOffset % TickSize;

    if (!LimitOffset)
      LimitOffset = TickSize << 1;

    const uint StopLevel = (uint)SymbolInfoInteger(Symb, SYMBOL_TRADE_STOPS_LEVEL) << 1;

    if (LimitOffset < StopLevel)
    {
      LimitOffset = StopLevel;
      LimitOffset += TickSize - (LimitOffset % TickSize);
    }

    TICKET_TYPE Ticket = 0;

    if (DemoAccount &&
      // https://www.mql5.com/ru/forum/191606/page5#comment_8158457
      (Ask - CorrectPrice(Symb, Ask) < Point / 2) && (Bid - CorrectPrice(Symb, Bid) < Point / 2) && (Last - CorrectPrice(Symb, Last) < Point / 2))
    {
      ERROR_LOG(OrderSend(Symb, OP_BUY, MinLot, Ask, 1000, 0, 0, __FUNCTION__ + "_Position", 0, 0, INT_MIN))

      if ((Ticket > 0) && OrderSelect(Ticket, SELECT_BY_TICKET))
      {
        ERROR_LOG(OrderModify(OrderTicket(), OrderOpenPrice(), 0, CorrectPrice(Symb, Bid + LimitOffset * Point), 0, INT_MIN))

        ERROR_LOG(OrderClose(OrderTicket(), OrderLots(), Bid, 1000, INT_MIN))
      }
    }

    double PriceMin = SymbolInfoDouble(Symb, SYMBOL_SESSION_PRICE_LIMIT_MIN);

    if (PriceMin < TickSize * Point)
      PriceMin = CorrectPrice(Symb, MathMax(Ask - LimitOffset * Point, TickSize * Point));

    if (PriceMin < Ask)
    {
      ERROR_LOG(OrderSend(Symb, OP_BUYLIMIT, MinLot, CorrectPrice(Symb, PriceMin), 1000, 0, 0, __FUNCTION__ + "_Order", 0, 0, INT_MIN))

      if ((Ticket > 0) && OrderSelect(Ticket, SELECT_BY_TICKET))
      {
        ERROR_LOG(OrderDelete(OrderTicket(), INT_MIN))

        if (!Ticket && !DemoAccount)
        {
          Alert("OrderDelete " + (string)OrderTicket() + " ERROR, Check Now!");

          ExpertRemove();

          return;
        }
      }
    }

    ERROR_LOG(OrderSend(Symb, OP_BUYSTOP, MinLot, CorrectPrice(Symb, MathMax(Ask, Last) + LimitOffset * Point), 1000, 0, 0, __FUNCTION__ + "_Order", 0, 0, INT_MIN))

    if ((Ticket > 0) && OrderSelect(Ticket, SELECT_BY_TICKET))
    {
      ERROR_LOG(OrderDelete(OrderTicket(), INT_MIN))

      if (!Ticket && !DemoAccount)
      {
        Alert("OrderDelete " + (string)OrderTicket() + " ERROR, Check Now!");

        ExpertRemove();

        return;
      }
    }
  }
}

void OnStart()
{
  if ((ENUM_ACCOUNT_TRADE_MODE)AccountInfoInteger(ACCOUNT_TRADE_MODE) == ACCOUNT_TRADE_MODE_DEMO) // ,    
  {
    Alert(TOSTRING(AccountInfoString(ACCOUNT_SERVER)) + TOSTRING(TerminalInfoInteger(TERMINAL_PING_LAST)));

    int Cycle = 0;

    while (!IsStopped() && (bool)(++Cycle))
    {
      for (int i = SymbolsTotal(true) - 1; (i >= 0) && (!IsStopped()); i--)
      {
        Comment(TOSTRING(Cycle) + TOSTRING(i) + SymbolName(i, true));

        OrderSendTest(SymbolName(i, true));
//    OrderSendTest();
      }
    }

    Comment("");
  }
}